/*
 * Copyright (C) 2024 Linux Studio Plugins Project <https://lsp-plug.in/>
 *           (C) 2024 Vladimir Sadovnikov <sadko4u@gmail.com>
 *
 * This file is part of lsp-dsp-lib
 * Created on: 27 нояб. 2024 г.
 *
 * lsp-dsp-lib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * lsp-dsp-lib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with lsp-dsp-lib. If not, see <https://www.gnu.org/licenses/>.
 */


#include <lsp-plug.in/common/types.h>
#include <lsp-plug.in/test-fw/utest.h>
#include <lsp-plug.in/test-fw/FloatBuffer.h>

namespace lsp
{
    namespace generic
    {
        void depan_eqpow(float *dst, const float *l, const float *r, float dfl, size_t count);
    }

    IF_ARCH_X86(
        namespace sse
        {
            void depan_eqpow(float *dst, const float *l, const float *r, float dfl, size_t count);
        }

        namespace avx
        {
            void depan_eqpow(float *dst, const float *l, const float *r, float dfl, size_t count);
            void depan_eqpow_fma3(float *dst, const float *l, const float *r, float dfl, size_t count);
        }

        namespace avx512
        {
            void depan_eqpow(float *dst, const float *l, const float *r, float dfl, size_t count);
        }
    )

    IF_ARCH_ARM(
        namespace neon_d32
        {
            void depan_eqpow(float *dst, const float *l, const float *r, float dfl, size_t count);
        }
    )

    IF_ARCH_AARCH64(
        namespace asimd
        {
            void depan_eqpow(float *dst, const float *l, const float *r, float dfl, size_t count);
        }
    )

    typedef void (* depan_t) (float *dst, const float *l, const float *r, float dfl, size_t count);
}

UTEST_BEGIN("dsp.pan", depan_eqpow)

    void call(const char *label, size_t align, depan_t func1, depan_t func2)
    {
        if (!UTEST_SUPPORTED(func1))
            return;
        if (!UTEST_SUPPORTED(func2))
            return;

        UTEST_FOREACH(count, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
                32, 33, 37, 48, 49, 64, 65, 0x3f, 100, 999, 0x1fff)
        {
            for (size_t mask=0; mask <= 0x07; ++mask)
            {
                printf("Testing %s on input buffer of %d numbers, mask=0x%x...\n", label, int(count), int(mask));

                FloatBuffer src1(count, align, mask & 0x01);
                FloatBuffer src2(count, align, mask & 0x02);
                FloatBuffer dst1(count, align, mask & 0x04);
                FloatBuffer dst2(dst1);

                // Call functions
                func1(dst1, src1, src2, 0.5f, count);
                func2(dst2, src1, src2, 0.5f, count);

                UTEST_ASSERT_MSG(src1.valid(), "Source buffer 1 corrupted");
                UTEST_ASSERT_MSG(src2.valid(), "Source buffer 2 corrupted");
                UTEST_ASSERT_MSG(dst1.valid(), "Destination buffer 1 corrupted");
                UTEST_ASSERT_MSG(dst2.valid(), "Destination buffer 2 corrupted");

                // Compare buffers
                if (!dst1.equals_absolute(dst2, 1e-5))
                {
                    src1.dump("src1");
                    src2.dump("src2");
                    dst1.dump("dst1");
                    dst2.dump("dst2");
                    UTEST_FAIL_MSG("Output of functions for test '%s' differs", label);
                }
            }
        }
    }

    UTEST_MAIN
    {
        #define CALL(func, align) \
            call(#func, align, generic::depan_eqpow, func)

        IF_ARCH_X86(CALL(sse::depan_eqpow, 16));
        IF_ARCH_X86(CALL(avx::depan_eqpow, 32));
        IF_ARCH_X86(CALL(avx::depan_eqpow_fma3, 32));
        IF_ARCH_X86(CALL(avx512::depan_eqpow, 64));
        IF_ARCH_ARM(CALL(neon_d32::depan_eqpow, 16));
        IF_ARCH_AARCH64(CALL(asimd::depan_eqpow, 16));
    }

UTEST_END


